module Utils exposing (..)

import Dict exposing (Dict)
import Dict.Extra
import Http
import Countries
import Statistics

-- Case insensitive comparison of strings.
icompare : String -> String -> Bool
icompare str other =
    String.toLower str == String.toLower other


-- List of countries (with some override w.r.t. ISO Standard).
countries : List Countries.Country
countries =
    let
        missing : List Countries.Country
        missing =
            [ Countries.Country "Kosovo" "XK" "🇽🇰" ]
    in
    Countries.all ++ missing

-- Dict of countries created from list (access in almost O(1)).
idxCountries : Dict String Countries.Country
idxCountries =
    countries
        |> List.map (\entry -> ( entry.code, entry ))
        |> Dict.fromList

countryNone : Countries.Country
countryNone =
    Countries.Country "??" "??" "??"

-- Get flag of ISO as Unicode String (fallback is country name).
flag : String -> String
flag iso =
    case lookupCode iso of
        Nothing ->
            "#" ++ iso
        Just entry ->
            entry.flag

-- Look up country entry (only case-insensitive matching) using given ISO code.
-- Note: On failure this will return a dummy entry with ??.
lookupCountry : String -> Countries.Country
lookupCountry country =
    -- We DO NOT use Countries.searchCustom here, as it filters with String.contains,
    -- which leads to problems for e.g. Mali and Somaliland.
    let
        synonyms : Dict String String
        synonyms =
            Dict.fromList
                [ ( "Hong Kong S.A.R.", "CN" )
                , ( "Former Yugoslav Macedonia", "MK" )
                , ( "Former Yugoslav Republic of Macedonia", "MK" )
                , ( "Venezuela", "VE" )
                , ( "United Kingdom", "GB" )
                , ( "Palestinian Authority", "PS" )
                , ( "Rumania", "RO" )
                , ( "Iran", "IR" )
                , ( "Iran, Islamic Republic of", "IR" )
                , ( "Vietnam", "VN" )
                , ( "USA", "US" )
                , ( "Tanzania", "TZ" )
                , ( "Tanzania,UND Republic", "TZ" )
                , ( "Taiwan", "TW" )
                , ( "Korea, Democratic People's Republic of", "KP" )
                , ( "South Korea", "KR" )
                , ( "North Korea", "KR" )
                , ( "Cote d'Ivoire", "CI" )
                , ( "Cote Ivoire", "CI" )
                , ( "Laos", "LA" )
                , ( "Syria", "SY" )
                , ( "Czech Republic", "CZ" )
                , ( "Congo (Democratic Republic)", "CD" )
                , ( "Congo, The Democratic Republic of the", "CD" )
                , ( "Congo, Repub. Democ.", "CD" )
                , ( "Congo (Brazzaville)", "CG" )
                , ( "Macao S.A.R.", "MO" )
                , ( "TÜRKIYE", "TR" )
                , ( "Bolivia", "BO" )
                , ( "Moldova", "MD" )
                , ( "Cape Verde", "CV" )
                , ( "Palestine", "PS" )
                , ( "Libyan Arab Jamahiriya", "LY" )
                , ( "Brunei", "BN" )
                , ( "Guinea Bissau", "GW" )
                , ( "Holy See (Vatican City State)", "VA" )
                , ( "N. Cyprus", "CY" )
                , ( "Somaliland", "SO" )
                , ( "West Bank", "PS" )
                ]
            |> Dict.Extra.mapKeys (\name -> String.toLower name)
        found : Maybe Countries.Country
        found =
            case Dict.get (String.toLower country) synonyms of
                Nothing -> 
                    countries
                        |> List.filter
                            (\entry -> icompare country entry.name)
                        |> List.head
                Just code ->
                    lookupCode code
    in
    Maybe.withDefault countryNone found

-- Looks up country entry for given ISO code.
lookupCode : String -> Maybe Countries.Country
lookupCode iso =
    Dict.get iso idxCountries

-- Checks whether given ISO Country Code is known.
isCodeValid : String -> Bool
isCodeValid iso =
    lookupCode iso == Nothing |> not 

-- Converts ISO Country Code to Country Name.
codeToName : String -> String
codeToName iso =
    lookupCode iso
        |> Maybe.map .name
        |> Maybe.withDefault countryNone.name

buildErrorMessage : Http.Error -> String
buildErrorMessage httpError =
    case httpError of
        Http.BadUrl message ->
            message

        Http.Timeout ->
            "Server is taking too long to respond. Please try again later."

        Http.NetworkError ->
            "Unable to reach server."

        Http.BadStatus statusCode ->
            "Request failed with status code: " ++ String.fromInt statusCode

        Http.BadBody message ->
            message


mean : List Float -> Float
mean vals =
    List.sum vals / (List.length >> toFloat) vals

-- Returns lower left and upper right corners of the bounding box
-- which surrounds the given points.
bbox : List ( Float, Float ) -> Maybe ( ( Float, Float ), ( Float, Float ) )
bbox points =
    let
        xExtent : Maybe ( Float, Float )
        xExtent = points |> List.map Tuple.first |> Statistics.extent

        yExtent : Maybe ( Float, Float )
        yExtent = points |> List.map Tuple.second |> Statistics.extent

        mapFn : ( Float, Float ) -> ( Float, Float ) -> ( ( Float, Float ), ( Float, Float ) )
        mapFn ( xmin, xmax ) ( ymin, ymax ) =
            ( ( xmin, ymin ), ( xmax, ymax ) )
    in
    Maybe.map2 mapFn xExtent yExtent

-- Returns the center of the given points.
centroid : List ( Float, Float ) -> Maybe ( Float, Float )
centroid p =
    case p of
        [] -> Nothing
        points -> 
            Just
            ( points |> List.map Tuple.first |> mean
            , points |> List.map Tuple.second |> mean
            )

-- Triangle Formula. Works only if polygon is not self-intersecting.
-- Requires the points of a polygon to be in (counter)clockwise order!
-- See also https://en.wikipedia.org/wiki/Shoelace_formula.
area : List ( Float, Float ) -> Float
area polygon =
    case polygon of
        [] -> 0
        [ _ ] -> 0
        [ _, _ ] -> 0
        ( x0, y0 ) :: next :: rest ->
            let
                triangleArea : ( Float, Float ) -> ( Float, Float ) -> Float
                triangleArea ( xs, ys ) ( xt, yt ) =
                    0.5 * abs (x0 * (ys - yt) + xs * (yt - y0) + xt * (y0 - ys))
            in
            List.map2 triangleArea (next :: rest) rest |> List.sum

-- Euclidian Distance between two points.
distance : ( Float, Float ) -> ( Float, Float ) -> Float
distance ( x, y ) ( x0, y0 ) =
    ( x - x0 ) ^ 2 + ( y - y0 ) ^ 2 |> sqrt 

-- Maximum possible radius of a circle at center
-- to all the other given surrounding points.
mpr : List ( Float, Float ) -> ( Float, Float ) -> Float
mpr points center =
    points |> List.map (distance center) |> List.minimum |> Maybe.withDefault 0